import waba.fx.*;       // Graphics
//import waba.io.*;
import waba.sys.*;      // Convert
import waba.ui.*;
import waba.util.*;     // Vector

import extra.util.Maths;  // note: this needs wextra version w/ cos public&fixed

// StringUtil

/**
 * @author <A HREF="mailto:weyer@kagi.com">Steve Weyer</A>
 * Copyright (c) 2001, S. Weyer. All right reserved.
 * @version waba 1.0
 * !!!a 08 Feb 2001
**/
public class TurtleArea extends Control { // adapted from Scribble(DrawArea)
  // otherwise object heap problems for complex drawings,
  // but since erase after List selection, shouldn't be a problem
  final boolean SAVE_PT = false;

  final int INIT_PT_LEN = 100;
  public final static float RAD_PER_DEG = Maths.PI / 180.0f;  //0.01745329252 == 3.14159265359/180.0
  public final static int PEN_UP = 0;
  public final static int PEN_DOWN = 1; // ?? colors, width(only 1 for drawLine)
  final int NEWT_SIZE = 2;
  final int NEWT_SIZE2 = 2 * NEWT_SIZE;

  Graphics drawg = null;
  int iX, iY;
  float fX, fY;
  int originX, originY;
  Vector pointVec;

  boolean visible;
  int penWidth;
  int angle;
  Turtle turtle;  // main window

  public TurtleArea(boolean pen, boolean dvis) { // Turtle turt) {
    angle = 0;
    penWidth = (pen) ? PEN_DOWN : PEN_UP;
    visible = dvis;
    //turtle = turt;
    if (SAVE_PT)
      pointVec = new Vector(INIT_PT_LEN);
    // init others in setRect, initDraw
  }

  public void setRect(int x, int y, int width, int height) {
    super.setRect(x,y,width,height);
    originX = width / 2; originY = height / 2;
    fX = originX; fY = originY;
  }

  public void onPaint(Graphics g) {
    // erase draw area in white
    g.setColor(255, 255, 255);
    g.fillRect(0, 0, this.width, this.height);

    // draw lines
    g.setColor(0, 0, 0);
    iX = originX; iY = originY;
    if (SAVE_PT) {
      Point pt;
      int inewX = iX, inewY = iY;
      for (int i=0; i < pointVec.getCount(); i++) {
        pt = (Point) (pointVec.get(i));
        inewX = pt.x; inewY = pt.y;
        if (pt.pen == PEN_DOWN)
          g.drawLine(iX, iY, inewX, inewY);
        iX = inewX; iY = inewY;
      };
    };
    fX = iX; fY = iY;
    if (drawg == null) {
      drawg = createGraphics();
      drawg.setClip(0, 0, this.width, this.height);
    };
    if (visible)
      createNewt(); // show
  }

  public void erase() {
    if (SAVE_PT)
      pointVec = new Vector(INIT_PT_LEN);
    home();
    repaint();
    }
/*
  public void onEvent(Event ev) {
    if (ev.type == PenEvent.PEN_DOWN) {
      int inewX = ((PenEvent)ev).x;
      int inewY = ((PenEvent)ev).y;
      int dx = inewX - iX, dy = iY - inewY; // flip y
      if (dx == 0 && dy == 0)
        return;
      goPt_View (inewX, inewY, penWidth); // just go there
      // for fun, display Dist and Deg
      int newAng =
        (dx == 0) ? (dy > 0) ? 0 : 180
        //90(0),45(45),0(90),-45(135),-90(180),-135(215),-180(270),135(-45)
        : 90 - (int) (Maths.atan2(dy,dx)/RAD_PER_DEG);
      if (newAng < 0) // e.g. 135->-45->315
        newAng = newAng + 360;
      turtle.setDistDeg((int) (Maths.sqrt(dx*dx + dy*dy)), newAng);
    };
  }
*/

// GO/MOVE

  public void home() {
    angle = 0;
    goPt_View(originX, originY, PEN_UP);
  }
/*
  public void setHome(int dx, int dy) {
    angle = 0;
    goPt_View(
      originX = (this.width / 2) + dx,
      originY = (this.height / 2)- dy,
      PEN_UP);
  }
*/
  public void go(float dist) {
    goPt_1(0, dist, penWidth);
  }

  public void goPt(float x, float y) { // relative
    goPt_1(x, y, penWidth);
  }

  public void move(float dist) {
    goPt_1(0, dist, PEN_UP);
  }

  public void movePt(float x, float y) { // relative
    goPt_1(x, y, PEN_UP);
  }

  public void goPt_1(float dx, float dy, int newWid) {
    // check common cases to avoid sin/cos
    if (dx == 0.0f && dy == 0.0f)
      return;
    switch (angle) {
      case 0:
        goPt_View(fX + dx, fY - dy, newWid);
        break;
      case 90:
        goPt_View(fX + dy, fY + dx, newWid);
        break;
      case 180:
        goPt_View(fX - dx, fY + dy, newWid);
        break;
      case 270:
        goPt_View(fX - dy, fY - dx, newWid);
        break;
      default:
        float dist, rads;
        if (dx == 0.0f) {
          dist = dy;
          rads = angle * RAD_PER_DEG;
          }
        else if (dy == 0.0f) {
          dist = dx;
          rads = ((angle >= 270) ? angle-270 : angle+90) * RAD_PER_DEG;
          }
        else {
          dist = Maths.sqrt(dx*dx + dy*dy); //hypot(dx,dy);
          rads = Maths.atan2(dy,dx);
          };
        goPt_View(fX + (dist * Maths.sin(rads)), fY - (dist * Maths.cos(rads)), newWid);
    };
  }

  public void goPt_View (float fnewX, float fnewY, int newWid) {
    int inewX = (int) (0.5f + fnewX), inewY = (int) (0.5f + fnewY);

    if (visible)
      createNewt(); // hide

    if (newWid != PEN_UP) {	// draw
      if (SAVE_PT && penWidth != newWid)
        pointVec.add(new Point(iX,iY,PEN_UP));
      drawg.drawLine(iX,iY, inewX,inewY);
    };
    if (SAVE_PT)
      pointVec.add(new Point(inewX, inewY, newWid));
    fX = fnewX; fY = fnewY;
    iX = inewX; iY = inewY;

    if (visible)
      createNewt(); // show
  }

// TURN

  public void turn(int ang) {
    if (ang != 0)
      turnTo(angle + ang);
  }

  public void turnTo(int ang) {
    // if (visible) createNewt(); // no need since it can't turn
    if (ang != angle) {
      ang %= 360;
      angle = (ang < 0) ? 360+ang : ang;
    };
    // if (visible) createNewt(); // no need since it can't turn
  }

// PEN

  public void pen(boolean pen) {
    //pen((pen) ? PEN_DOWN : PEN_UP);
    penWidth = (pen) ? PEN_DOWN : PEN_UP;
  }

  // uncomment the following for cleaner user methods
/*
  public void penDown() {
    pen(PEN_DOWN);
  }
  public void penUp() {
    pen(PEN_UP);
  }
  public void pen(int newWid) {
    //if (visible) createNewt(); // ??
    penWidth = newWid;
    //if (visible) createNewt(); // ??
  }
*/

// "NEWT"

  public void setVis(boolean vis) {
    if (visible)
      createNewt(); // hide

    if (visible = vis)
      createNewt(); // show
  }

  public void createNewt() {
    drawg.drawCursor(iX-NEWT_SIZE, iY-NEWT_SIZE, NEWT_SIZE2, NEWT_SIZE2);
  }

// EXAMPLES

  public boolean doEval(String methodName, int times, int dist, int deg) {
//System.out.println(expr + " " + times + ',' + dist + ',' + deg);
    if (methodName.equals("go"))            // Dist button
      go(dist);
    else if (methodName.equals("turn"))     // Deg button
      turn(deg);
    else if (methodName.equals("poly2"))    // Times button
      poly2(times,dist,deg,0);
    else if (methodName.equals("poly")) { // Eval
      movePt(-dist/2, -dist/2); // "center"
      poly2(times,dist,360/times,0);
    }
    else if (methodName.equals("squiral"))
      squiral(times, dist, deg);
    else if (methodName.equals("polySpi"))
      polySpi(times, dist, deg);
    else if (methodName.equals("cCurve"))
      cCurve(times, dist, deg);
    else if (methodName.equals("dCurve"))
      dCurve(times, dist, deg);
    else if (methodName.equals("hilbert"))
      hilbert(times, dist, deg);
    else if (methodName.equals("gasket"))
      gasket(times, dist, deg);
    else if (methodName.equals("carpet"))
      carpet(times, dist, deg);
    else if (methodName.equals("fractal"))
      fractal(times, dist, deg);
    else if (methodName.equals("tree"))
      tree(times, dist, deg);
    else {
      Sound.beep(); // ?? error
      return false;
    };
    return true;
  }

  public void poly2(int nsides, int dist, int deg, int deg2) {
    for (int i=0; i < nsides; i++) {
      go(dist); turn(deg);
    };
    turn(deg2);
  }

  public void squiral(int times, int dist, int deg) {
    int deg2 = 360 / times, sides = 360 / deg;
    for (int i=0; i < times; i++)
      poly2(sides,dist,deg,deg2);
  }


  public void cCurve(int times, int dist, int deg) {
    movePt(0, -dist*(times-1)); // "center"
    cCurve2(times,dist,deg);
  }
  public void cCurve2(int times, int dist, int deg) {
    if (times > 0) {
      cCurve2(--times, dist, deg+45);
      cCurve2(  times, dist, deg-45);
    }
    else {
      turnTo(deg); go(dist);
    };
  }

  public void dCurve(int times, int dist, int deg) {
    movePt(0, -dist*(times-1)); // "center"
    dCurve2(times, dist, deg, 45);
  }
  public void dCurve2(int times, int dist, int deg, int ddeg) {
    if (times > 0) {
      dCurve2(--times, dist, deg+ddeg,  45);
      dCurve2(  times, dist, deg-ddeg, -45);
    }
    else {
      turnTo(deg); go(dist);
    };
  }

  public void polySpi(int times, int dist, int deg) {
    movePt(0, -dist/2); // "center"
    for (int i=0; i < times; i++) {
      go(dist++); turn(deg);
    };
  }

  public void hilbert(int times, int dist, int deg) {
    movePt(originX/2, dist-originY); // "center"
    hilbert2(times,dist,deg);
  }
  public void hilbert2(int times, int dist, int deg) {
    if (times >= 1 && times <= 5) {
      turn(-deg);
      hilbert2(--times, dist, -deg);
      go(dist); turn(deg);
      hilbert2(times, dist, deg);
      go(dist);
      hilbert2(times, dist, deg);
      turn(deg); go(dist);
      hilbert2(times, dist, -deg);
      turn(-deg);
    };
  }

  public void gasket(int times, int dist, int deg) {
    // Serpenski Gasket. size should be 2^n, deg=120
    movePt(0, -dist/2); // "center"
    gasket1(times,dist,deg);
  }
  public void gasket1(int times, int dist, int deg) {
    for (int i=0; i < 360 / deg; i++) {
      gasket2(times, dist, deg);
      turn(deg);
    };
  }
  public void gasket2(int times, int dist, int deg) {
    if (times > 0 && dist >= 2) {
      dist = dist / 2;
      times--;
      gasket2(times, dist, deg);
      turn(-deg);
      gasket1(times,dist, deg);
      turn(deg);
      gasket2(times, dist, deg);
    }
    else
      go(dist);
  }

  public void carpet(int times, int dist, int deg) {
    movePt(-dist/3, -dist/3); // "center"
    carpet1(times,dist,deg);
  }
  public void carpet1(int times, int dist, int deg) {
    // Serpenski Carpet. size should be 3^n
    // note: level >= 3
    for (int i=0; i < 360 / deg; i++) {
      carpet2(times, dist, deg);
      turn(deg);
    };
  }
  public void carpet2(int times, int dist, int deg) {
    if (times > 0 && dist >= 3) {
      int dist3 = dist / 3;
      times--;
      for (int i=0; i < 3; i++)
        carpet2(times,dist3,deg);
      penWidth = PEN_UP; //pen(PEN_UP); //penUp();
      go(-dist-dist3); turn(-deg);
      go(dist3); turn(-deg);
      penWidth = PEN_DOWN; //pen(PEN_DOWN); //penDown();

      carpet1(times,dist3,deg);

      penWidth = PEN_UP; //pen(PEN_UP); //penUp();
      go(-dist);
      penWidth = PEN_DOWN; //pen(PEN_DOWN); //penDown();

      carpet1(times,dist3,deg);

      penWidth = PEN_UP; //pen(PEN_UP); //penUp();
      go(-dist3); turn(deg);
      go(-dist3); turn(deg);
      penWidth = PEN_DOWN; //pen(PEN_DOWN); //penDown();
    }
    else
      go(dist);
  }

  public void tree(int times, int dist, int deg) {
    movePt(0, -originY); // "center" (at bottom)
    tree2(times,dist,deg);
  }
  public void tree2(int times, int dist, int deg) {
    if (times > 0) {
      int dist23 = (dist*2) / 3;
      go(dist); turn(-deg);
      tree2(--times, dist23, deg);
      turn(deg*2);
      tree2(  times, dist23, deg);
      turn(-deg); go(-dist);
    };
  }

  public void fractal(int times, int dist, int deg) {
    movePt(-dist/4, -dist/2); // move from home
    for (int i=0; i < 360/deg; i++) {
      fractal2(times, dist, deg/2);
      turn(deg);
    };
  }
  public void fractal2(int times, int dist, int deg) {
    if (times > 0) {
      dist = dist / 3;
      times--;
      fractal2(times,dist,deg);
      turn(-deg);
      fractal2(times,dist,deg);
      turn(deg*2);
      fractal2(times,dist,deg);
      turn(-deg);
      fractal2(times,dist,deg);
    }
    else
      go(dist);
  }
}

/* // only needed if SAVE_PT
class Point {
  public int x;
  public int y;
  public int pen;

  public Point(int px, int py, int ppen) {
    x = px;
    y = py;
    pen = ppen;
  }
}
*/